//---------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <iostream>
#include <iomanip>
using namespace std;

//---------------------------------------------------------------------------

typedef struct __Cidade {

  int x;          // coordenada no eixo x do plano
  int y;          // coordenada no eixo y do plano
  int demanda;    // demanda da cidade

} Cidade;

//---------------------------------------------------------------------------

typedef struct __Candidato {

  int position;
  int cidade;
  double economia;

} Candidato;

//---------------------------------------------------------------------------

typedef struct __Movimento {

  int p, q;

} Movimento;

//---------------------------------------------------------------------------

Cidade *ler_arquivo(char nome_arquivo[], int *n, int *capacidade);
int *cria_vetor(int tamanho);
int tamanho_vetor(int *s_star, int tam);
double **cria_matriz(int nlinhas, int ncolunas);
void inicializa_vetor(int *vetor, int tamanho);
void inicializa_matriz(double **matriz, int nlinhas, int ncolunas);
void calcula_distancia(double **matriz, Cidade *cidades, int nlinhas, int ncolunas);
double calcular_fo(int *s, double **dist, int size);
double calcular_fa(int *s, int tam, double penalidade, int capacidade, Cidade *cidades, double **dist);
Movimento gerarMovimento(int *s, int tamanho, Cidade cidades, double **distancia, int capacidade, int penalidade);
void s_swap(int *s, int p, int q);
bool movimentoTabu(Movimento mov, Movimento *listaTabu, int Tsize, int *s);
int resize(Movimento *listaTabu, int Tsize, int posIns, int Tmin, int Tmax);
int function_comp(const void *i, const void *j);
void quick_sort(Candidato *lc, int size);
void construir_solucao_inicial(int n, double **dist, int *s, Cidade *cidades, int cap, double alfa);
int *buscaTabu(int *s, double fo_star, int tamanho, double **distancia, Cidade *cidades, int n, int capacidade, int penalidade, int BTmax, int Tmin, int Tmax, double Fmin);

//---------------------------------------------------------------------------

int main()
{

  int n;                    // nmero de cidades + depsito
  double **distancia;       // matriz de distncias entre as cidades
  int capacidade;           // capacidade da frota
  int *s;                   // vetor soluo
  int *s_star;              // vetor melhor soluo encontrada
  int *s_aux;               // vetor auxiliar
  int tamanho;              // tamanho do vetor soluo
  int seed;                 // semente da funo aleatria
  double fo = 0.0;          // funo objetivo
  double fo_star = 0.0;     // melhor funo objetivo
  Cidade *cidades;          // vetor de cidades

  /* Parametros da Metaheurstica GRASP */
    int iter=0;                       // mximo de iteraes para gerar a soluo inicial ( Filtro )
    int maxGRASP = 1;                 // mximo de iteraes para o GRASP
    int iterSo = 1100;                // mximo de geraes da soluo inicial
    double alfaGRASP = 0.03;          // parmetro alfa para o GRASP (fator aleatoriedade)
    double ini_time = time(NULL);     // incio do processamento
    double end_time = 300;            // tempo para o trmino do processamento em segundos
    double Fmin = 524.61;             // melhor valor da literatura
    int BTmax = 75000;                // mximo de iteraes sem melhora com o busca tabu
    double penalidade = 800;          // penalidade por unidade que excede a capacidade
    int Tmin = 0;                     // tamanho mnimo da lista tabu
    int Tmax = 10;                    // tamanho mximo da lista tabu
  /* ---------------------------------- */

  /* Procedimento inicial */
    cidades = ler_arquivo("c50.txt", &n, &capacidade);
    tamanho = 2*n+1;                                    // tamanho mximo do vetor soluo

    s = cria_vetor(tamanho);                            // cria vetor soluo
    s_star = cria_vetor(tamanho);                       // cria vetor melhor soluo
    distancia = cria_matriz(n+1, n+1);                  // aloca a matriz dinamicamente
    inicializa_matriz(distancia, n+1, n+1);             // inicializa matriz
    calcula_distancia(distancia, cidades, n+1, n+1);    // calcula distancia entre as cidades
  /* -------------------- */

  /* GRASP - Greedy Randomized Adaptive Search Procedure */

    fo_star = 10000.0;                                  // fo inicializada com +infinito

    while ( fo_star > Fmin && iter++ < maxGRASP && time(NULL)-ini_time < end_time) {

      seed = (unsigned) time(NULL);
      srand( seed );                   // gerar semente para funo random()
      cout << "Semente: " << seed << "\n";

      /* Fase de construo da soluo inicial */
        tamanho = 2*n+1;                                // tamanho mximo do vetor soluo
        for (int iterF=0; iterF<iterSo; iterF++) {
          inicializa_vetor(s, 2*n+1);                   // inicializa vetor soluo
          construir_solucao_inicial(n, distancia, s, cidades, capacidade, alfaGRASP); // gera soluo inicial
          fo = calcular_fo(s, distancia, tamanho);
          if ( fo < fo_star ) {
            fo_star = fo;                               // salva melhor fo
            inicializa_vetor(s_star, 2*n+1);            // inicializa vetor melhor soluo
            for (int m=0; m<tamanho; m++) {             // salva melhor soluo
              s_star[m] = s[m]; }
          }
        }
        tamanho = tamanho_vetor(s_star, 2*n+1);         // reconfigura o tamanho do vetor soluo
        free(s);                                        // libera o vetor soluo

      /* Refinamento com busca local :: Busca Tabu */
        s = buscaTabu(s_star, fo_star, tamanho, distancia, cidades, n, capacidade, penalidade, BTmax, Tmin, Tmax, Fmin);

      fo = calcular_fo(s, distancia, tamanho);

      if ( fo <= fo_star ) {
        fo_star = fo;                                   // salva melhor fo
        inicializa_vetor(s_star, 2*n+1);                // inicializa vetor melhor soluo
        for (int m=0; m<2*n+1; m++) {                   // salva melhor soluo
          s_star[m] = s[m]; }
      }

      cout << "Solucao: s=[ "; for (int e=0; e<tamanho; e++) { cout << s_star[e] << " "; } cout << "]\n"; cout << "Funcao objetivo: " << fo_star << " Funo de avaliacao: " << calcular_fa(s_star, tamanho, penalidade, capacidade, cidades, distancia) << " Numero de cidades: \n";/**/

    } // fim for;

  /* ------------------ FIM DO GRASP ------------------- */

  int num = 0;
  for (int i=1; i<51; i++) {
    for (int j=0; j<tamanho; j++) {
      if ( s[j] == i ) {
        num++;
        break;
      }
    }
  }

  cout << "Solucao: s=[ "; for (int e=0; e<tamanho; e++) { cout << s_star[e] << " "; } cout << "]\n"; cout << "Funcao objetivo: " << fo_star << " Funo de avaliacao: " << calcular_fa(s_star, tamanho, penalidade, capacidade, cidades, distancia) << " Numero de cidades: " << num << "\n";/**/

  getchar();
  return 0;

} // main

//---------------------------------------------------------------------------

Cidade *ler_arquivo(char nome_arquivo[], int *n, int *capacidade)
{

  int i, x, y, d;
  FILE *arquivo;
  Cidade *cidades;

  arquivo=fopen(nome_arquivo, "r");
  if (!arquivo) {
    printf("No foi possvel abrir o arquivo [%s].\n", nome_arquivo); getchar();
    exit(1);
  }

  fscanf( arquivo, "%d", n );
  fscanf( arquivo, "%d", capacidade );

  cidades = (Cidade *) malloc((*n+1)*sizeof(Cidade));

  while (!feof(arquivo)) {
    fscanf( arquivo, "%d %d %d %d", &i, &x, &y, &d );
    cidades[i].x = x;
    cidades[i].y = y;
    cidades[i].demanda = d;
  }

  fclose( arquivo );
  return cidades;

} // ler_arquivo

//---------------------------------------------------------------------------

int *cria_vetor(int tamanho)
{

  int *vetor;

  vetor = (int *) malloc(tamanho*sizeof(int));
  if (!vetor) {
    printf("Memria insuficiente para criar vetor");
    exit(1);
  }
  return vetor;

} // cria_vetor

//---------------------------------------------------------------------------

int tamanho_vetor(int *solucao, int tam)
{

  int size=0;

  for (int i=0; i<tam-1; i++) {
    if ( solucao[i] == 0 && solucao[i+1] == 0 ) {
      size = i+1;
      break;
    }
  }

  return size;

} // tamanho_vetor

//---------------------------------------------------------------------------

double **cria_matriz(int nlinhas, int ncolunas)
{

  double **matriz;

  matriz = (double **) malloc(nlinhas*sizeof(double *));
  if (!matriz) {
    printf("Memria insuficiente para criar matriz!");
    exit(1);
  }
  for (int i=0; i<nlinhas; i++) {
    matriz[i] = (double *) malloc(ncolunas*sizeof(double));
    if (!matriz) {
      printf("Memria insuficiente para criar matriz!");
      exit(1);
    }
  }

  return matriz;

} // fim cria_matriz;

//---------------------------------------------------------------------------

void inicializa_vetor(int *vetor, int tamanho)
{

  for (int i=0; i<tamanho; i++) {
    vetor[i] = 0;
  }

} // inicializa_vetor

//---------------------------------------------------------------------------

void inicializa_matriz(double **matriz, int nlinhas, int ncolunas)
{

  for (int i=0; i<nlinhas; i++) {
    for (int j=0; j<ncolunas; j++) {
      matriz[i][j] = 0;
    }
  }

} // inicializa_matriz

//---------------------------------------------------------------------------

void calcula_distancia(double **matriz, Cidade *cidades, int nlinhas, int ncolunas)
{
  double yop = 0.0;   // calcula a diferenca entre as coordenadas y das cidades e as levam ao quadrado
  double xop = 0.0;   // calcula a diferenca entre as coordenadas x das cidades e as levam ao quadrado
  double xyop = 0.0;  // calcula a raiz quadrada entra xop e yop
  double x = 0.0;     // captura a coordenada x da cidade
  double y = 0.0;     // captura a coordenada y da cidade
  double x1 = 0.0;    // captura a coordenada x1 da cidade
  double y1 = 0.0;    // captura a coordenada y1 da cidade

   for (int i = 0; i < nlinhas; i ++) {
      x = cidades[i].x;
      y = cidades[i].y;
    for (int j = i+1; j < ncolunas; j++ ) {
      x1 = cidades[j].x;
      y1 = cidades[j].y;

      xop = (x - x1)*(x - x1);
      yop = (y - y1)*(y - y1);
      xyop = xop + yop;
      matriz[i][j] = sqrt(xyop);  // Considera-se dij = dji
      matriz[j][i] = matriz[i][j];
    }
  }

} // calcula_distancia

//---------------------------------------------------------------------------

double calcular_fo(int *s, double **dist, int size)
{

  int tam = size - 1;
  double fo=0.0;
  for (int i=0; i<tam; i++) {
    fo = fo + dist[(s[i])][(s[i+1])];
  }
  return (double)fo;

} // calcular_fo

//---------------------------------------------------------------------------

double calcular_fa(int *s, int tam, double penalidade, int capacidade, Cidade *cidades, double **dist)
{

  int w = 0;
  int w_exc = 0;
  double fo = 0.0;

  for (int m=0; m<tam; m++) {
    fo = fo + dist[(s[m])][(s[m+1])];
    w += cidades[(s[m])].demanda;
    if ( s[m] == 0 ) {
      if ( w > capacidade ) {
        w_exc += ( w-capacidade );
      }
      w = 0;
    }
  }

  return fo + (penalidade*w_exc);

} // calcular_fa

//---------------------------------------------------------------------------

Movimento gerarMovimento(int *s, int tamanho, Cidade *cidades, double **distancia, int capacidade, int penalidade)
{

  double fa, m_fa;
  Movimento mov; mov.p = 0; mov.q = 0;
  int petalaI, petalaII, ini_pI, ini_pII, fim_pI, fim_pII;

  int choiceI = ( random(100) > 50 ? 0 : 2 );
  int choice = ( random(100) > 70 ? 1 : choiceI );

  switch ( choice ) {

    case 0:
          fa = 0.0; m_fa = penalidade*10000.0;
          do { petalaI = random(tamanho); } while ( s[petalaI] == 0 );
          do { petalaII = random(tamanho); } while ( s[petalaII] == 0 );
          for (ini_pI=petalaI; ini_pI>=0; ini_pI--) { if (s[ini_pI]==0) break; }
          for (fim_pI=petalaI; fim_pI<tamanho; fim_pI++) { if (s[fim_pI]==0) break; }
          for (ini_pII=petalaII; ini_pII>=0; ini_pII--) { if (s[ini_pII]==0) break; }
          for (fim_pII=petalaII; fim_pII<tamanho; fim_pII++) { if (s[fim_pII]==0) break; }
          for (int i=ini_pI+1; i<fim_pI; i++) {
            for (int j=ini_pII+1; j<fim_pII; j++) {
              s_swap(s, i, j);
              fa = calcular_fa(s, tamanho, penalidade, capacidade, cidades, distancia);
              if ( fa < m_fa ) {
                m_fa = fa;
                mov.p = i; mov.q = j;
              }
              s_swap(s, i, j);
            }
          }
          break;
    case 1:
          fa = 0.0; m_fa = penalidade*10000.0;
          do { petalaI = random(tamanho); } while ( s[petalaI] == 0 );
          for (ini_pI=petalaI; ini_pI>=0; ini_pI--) { if (s[ini_pI]==0) break; }
          for (fim_pI=petalaI; fim_pI<tamanho; fim_pI++) { if (s[fim_pI]==0) break; }
          for (int i=ini_pI+1; i<fim_pI; i++) {
            for (int j=i+1; j<fim_pI; j++) {
              s_swap(s, i, j);
              fa = calcular_fa(s, tamanho, penalidade, capacidade, cidades, distancia);
              if ( fa < m_fa ) {
                m_fa = fa;
                mov.p = i; mov.q = j;
              }
              s_swap(s, i, j);
            }
          }
          break;
    case 2:
          fa = 0.0; m_fa = penalidade*10000.0;
          for (int i=1; i<tamanho-1; i++) {
            for (int j=i+1; j<tamanho-1; j++) {
              s_swap(s, i, j);
              fa = calcular_fa(s, tamanho, penalidade, capacidade, cidades, distancia);
              if ( fa < m_fa ) {
                m_fa = fa;
                mov.p = i; mov.q = j;
              }
              s_swap(s, i, j);
            }
          }
  }

  return mov;

} // gerarMovimento

//---------------------------------------------------------------------------

void s_swap(int *s, int p, int q)
{

  int a = s[q];
   s[q] = s[p];
   s[p] = a;

} // s_swap

//---------------------------------------------------------------------------

bool movimentoTabu(Movimento mov, Movimento *listaTabu, int Tsize, int *s)
{

  for (int i=0; i<Tsize; i++) {
    if ( listaTabu[i].p == mov.p && listaTabu[i].q == mov.q ) {
      s_swap(s, mov.q, mov.p);
      return true;
    }
  }
  return false;

} // movimentoTabu

//---------------------------------------------------------------------------

int resize(Movimento *listaTabu, int Tsize, int posIns, int Tmin, int Tmax)
{

  int newSize = random( Tmax - Tmin ) + Tmin;
  Movimento *LTdinamic = listaTabu;
  listaTabu = (Movimento *) malloc( newSize*sizeof(Movimento) );

  for (int m=0; m<newSize; m++) { listaTabu[m].p = 0; listaTabu[m].q = 0; }

  posIns--;

  for (int m=0; m<Tsize && m<newSize; m++) {
    if (posIns == -1) { posIns = Tsize - 1; }
    listaTabu[newSize-(1+m)] = LTdinamic[posIns];
    posIns--;
  }

  return newSize;

}

//---------------------------------------------------------------------------

int function_comp(const void *i, const void *j)
{

  Candidato *x = (Candidato *)i;
  Candidato *y = (Candidato *)j;

  if ( (x->economia) < (y->economia) ) return 1;
  if ( (x->economia) > (y->economia) ) return -1;
  return 0;

} // function_comp

//---------------------------------------------------------------------------

void quick_sort(Candidato *lc, int size)
{

  qsort( lc, size, sizeof(Candidato), function_comp );

} // quick_sort

//---------------------------------------------------------------------------

void construir_solucao_inicial(int n, double **d, int *s, Cidade *cidades, int cap, double alfa)
{

  int iniP=0;                 // incio da ptala
  int fimP=1;                 // fim da ptala
  int numC=n;                 // numero de candidatos
  int demanda=0;              // demanda da petala
  int i=0, j=0;               // variaveis auxiliares para cidades i, j
  double economia=0;          // economia

  int LCR = numC;             // 'lista' de candidatos restrita: intervalo [0, LCR]
  Candidato *c = (Candidato *) malloc( n*sizeof(Candidato) ); // cria 'lista' de candidatos
  for (int k=0; k<n; k++) { c[k].cidade = n-k; }              // inicializa 'lista' de candidatos

  int p_rand = 0;             // posio aleatria

  do {

    /* Incio da ptala */
      numC = numC - 1;
      p_rand = random(numC/2);
      s[fimP] = c[p_rand].cidade;
      c[p_rand] = c[numC];
      fimP = fimP+1;
      demanda=cidades[(s[iniP+1])].demanda;/**/

    /* Formao de uma ptala */
    while ( demanda < cap && numC > 0 ) {

      /* calcular as economias */
        for (int k=0; k<numC; k++) { // ini-for(1);
          i = s[iniP+1]; j = c[k].cidade;
            economia = d[0][i] + d[0][j] - d[i][j];
            c[k].economia = economia;
            c[k].position = iniP+1;

          i = s[fimP-1];
            economia = d[0][i] + d[0][j] - d[i][j];
            if (economia >= c[k].economia) {
              c[k].economia = economia;
              c[k].position = fimP;
            }
        } // end-for(1);
      /* --------------------- */

        quick_sort(c, numC);
        LCR = (int) numC*alfa;
        p_rand = random(LCR);

        demanda = demanda + cidades[ (c[p_rand].cidade) ].demanda;

      /* inserir cidade com a melhor economia */
        if ( demanda <= cap ) {
          if ( c[p_rand].position == iniP+1 ) { for (int q=fimP; q>iniP+1; q--) { s[q] = s[q-1]; } }
          s[(c[p_rand].position)] = c[p_rand].cidade;
          numC = numC-1;
          c[p_rand] = c[numC];
          fimP = fimP+1;
        }

      } // fim while() {};

    /* Fim da petala */
      iniP = fimP;
      fimP = fimP+1;

  } while ( numC > 0 ); // fim do{} while();

  free(c);    // libera vetor c ( vetor candidatos );

} // construir_solucao_inicial

//---------------------------------------------------------------------------

int *buscaTabu(int *s, double fo_star, int tamanho, double **distancia, Cidade *cidades, int n, int capacidade, int penalidade, int BTmax, int Tmin, int Tmax, double Fmin)
{

  /* PARAMETERS: s, fo_star, tamanho, distancia, cidades, n, capacidade, penalidade, BTmax, Tmax */

  int *s_star = cria_vetor(2*n+1);  // cria vetor melhor soluo
  int iter = 0;                     // contador de iteraes
  int bestIter = 0;                 // iterao mais recente que forneceu s'
  int counter = 0;                  // contador do nmero de iteraes com tamanho fixo da lista tabu
  double fa = 0.0;                  // funo avaliao
  int posIns = 0;                   // posio para insero na lista tabu
  int Tsize = Tmax;                 // tamanho da lista tabu
  Movimento *listaTabu;             // lista tabu
  Movimento mov;                    // movimento gerado

  listaTabu = (Movimento *) malloc( Tsize*sizeof(Movimento) );
  for (int m=0; m<2*n+1; m++) { s_star[m] = s[m]; }
  for (int m=0; m<Tsize; m++) { listaTabu[m].p = 0; listaTabu[m].q = 0; }

  while ( fo_star > Fmin && iter - bestIter < BTmax ) {

    iter++;

    if ( Tmin!=0 ) {
      counter++;
      if ( iter - bestIter > BTmax*0.4 ) {
        resize(listaTabu, Tsize, posIns, Tmin, Tmax);
        counter=0;
      }
    }

    do {

      mov = gerarMovimento(s, tamanho, cidades, distancia, capacidade, penalidade);
      s_swap(s, mov.q, mov.p);
      fa = calcular_fa(s, tamanho, penalidade, capacidade, cidades, distancia);

      if ( fa < fo_star ) break; // aceita o movimento se ocorrer melhora na melhor soluo

    } while ( movimentoTabu(mov, listaTabu, Tsize, s) );

    if ( posIns == Tsize ) posIns = 0;
    listaTabu[posIns] = mov; posIns++;

    if ( fa < fo_star ) {
      fo_star = fa;
      //cout << " OLD S_STAR = [ "; for (int m=0; m<tamanho; m++) { cout << s_star[m] << " "; }
      for (int m=0; m<tamanho; m++) { s_star[m] = s[m]; }
      bestIter = iter;
      cout << "\n NEW S_STAR = [ "; for (int m=0; m<tamanho; m++) { cout << s_star[m] << " "; }
      cout << "] FA: " << fa << " FO_STAR: " << fo_star << "\n";
    }

    //cout << "FA: " << fa << "\tFO_STAR: " << fo_star << /*"\ti: " << iter <<*/ "\tbI: " << bestIter << "\tiR: " << BTmax - (iter - bestIter);
    //if ( calcular_fo(s, distancia, tamanho) == fa ) { cout << " FO: " << calcular_fo(s, distancia, tamanho); } cout << "\n";

  }

  return s_star;

} // buscaTabu

//---------------------------------------------------------------------------

